/* +------------------------------------------------------------------------+
   |                                                                        |
   |                       Division de Burnikel et Ziegler                  |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

static ndouble un = 1;

/* Multiplication disymtrique, lb-1 <= la <= lb */
static void multiply(ndouble *c, ndouble *a, ndouble *b, int la, int lb, int n0) {

  karamul(c,a,b,la,n0);

  if (la < lb) {
    int i;
    ndouble r, b0=b[lb-1];

    for (i=r=0, c+=lb-1; i<la; i++,a++,c++) {
      r   += a[0]*b0 + c[0];
      c[0] = r & mask;
      r  >>= shift;
    }
    c[0] = r;
  }
}

/* c <- trunc(a/b), lg(a) = 2n, lg(b) = lg(q) = n                 */
/* si rem = 0 alors a <- a - high(b*c) sinon a <- a - b*c         */
/* si rem = 0 le quotient est trop grand d'au plus 2log(n) units */
/* le chiffre de tte de b doit avoir son bit de poids fort  1   */
void burnikel(int rem, ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  if (n <= n0) divn2(c,a,b,n);
  else {
    int q=n/2, p=n-q, i,j;
    ndouble *buff = getmem(n+1);
    zdouble r;

                                          /* premier chiffre        */
    if (compare(a+n+q,b+q,p) >= 0) {      /* blocs de tte gaux ?  */
      clearmem(c+q,p);                    /* => tronque le quotient */
      r = dec(a+n,b,n,n);
      while(r) {dec(c+q,&un,p,1); r += inc(a+q,b,n+p,n);}
    }
    else {                                /* sinon,                 */
      burnikel(1, c+q, a+2*q, b+q, p,n0); /* divise les poids forts */
      multiply(buff,b, c+q, q,p,n0);      /* retranche b0*c1        */
      r = dec(a+q, buff, n,n);
      while (r) {dec(c+q,&un, p,1); r += inc(a+q, b, n,n);}
    }
                                          /* deuxime chiffre       */
    if (compare(a+n,b+p,q) >= 0) {        /* blocs de tte gaux ?  */
      for (i=0; i<q; i++) c[i] = mask;    /* => tronque le quotient */
      clearmem(a+n,q);
      for (i=q; i>0;) {
        j = ((i <= n0) || (rem)) ? i : (i+1)/2;
        inc(a+n-j,b+n-i,j+1,j);
        i -= j;
      }
    }
    else burnikel(rem, c, a+p, b+p, q, n0); /* sinon divise */

    if (rem) {
      multiply(buff, c,b, q,p,n0);         /* retranche b0*c0  */
      r = dec(a, buff, n+1,n);
      while (r) {dec(c,&un, n,1); r += inc(a, b, n+1,n);}
    }

    freemem(buff);
  }
}

/* corrige la division sans reste */
void burnikel_correct(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  if (n > n0) {
    int q=n/2, p=n-q;
    ndouble *buff;
    zdouble r;

    burnikel_correct(c,a+p,b+p,q,n0);

    buff = getmem(n+1);
    multiply(buff, c,b, q,p,n0);
    r = dec(a, buff, n+1,n);
    while (r) {dec(c,&un, n,1); r += inc(a, b, n+1,n);}
    freemem(buff);
  }
}

                    /* +-----------------------+
                       |  Division sans reste  |
                       +-----------------------+ */

void burnikel_nr(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  if (n <= n0) divn2(c,a,b,n);
  else {

    /* effectue la division sans reste avec deux chiffres aprs la virgule */
    ndouble *a1 = getmem(2*n+4);
    ndouble *b1 = getmem(n+2);
    ndouble *c1 = getmem(n+2);

    movemem(a1+4,a,2*n); clearmem(a1,4);
    movemem(b1+2,b,n);   clearmem(b1,2);
    burnikel(0,c1,a1,b1,n+2,n0);

    /* si la partie dcimale est trop petite corrige la division */
    if (c1[1] == 0) burnikel_correct(c1,a1,b1,n+2,n0);

    /* tronque le rsultat */
    movemem(c,c1+2,n);

    freemem(a1); freemem(b1); freemem(c1);
  }
}

/* tests */
void burnsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = 2*n;
  *sb = n;
  *sc = -n;
  *sd = -2*n;
}

void burn0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  b[n-1] |= msb; a[2*n-1] %= b[n-1];
  movemem(d,a,2*n);
  burnikel_nr(c,d,b,n,n0);

}

void burn1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  b[n-1] |= msb; a[2*n-1] %= b[n-1];
  movemem(d,a,2*n);
  burnikel(1,c,d,b,n,n0);

}

int divnrchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, r;
  int res;

  e = getmem(2*n);
  karamul(e, b, c, n, n0);
  res = (compare(e,a,2*n) > 0);
  r = inc(e, b, 2*n, n);
  res |= ((r == 0) && (compare(e,a,2*n) <= 0));
  freemem(e);
  return(res);
}

int divchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, r;
  int res;

  e = getmem(2*n);
  karamul(e, b, c, n, n0);
  r = inc(e, d, 2*n, n);
  res = (compare(e,a,2*n)) || (compare(d,b,n) >= 0);
  freemem(e);
  return(res);
}
