/* +------------------------------------------------------------------------+
   |                                                                        |
   |                             Racine carre modulaire                    |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

static ndouble un = 1;


/* Entres
    1 <= n <= base
    a de longueur 2n, les deux bits de poids fort valent 0,1
    meth = 0 ou 1

  Sortie
  si meth = 1
    c <- 2*floor(sqrt(a)), de longueur n
    a <- a - (c/2)^2
  si meth = 0
    c <- 2*approx(sqrt(a)) de longueur n
    la racine carre exacte est comprise entre c-2 et c
*/
void msqrt(int meth, ndouble *c, ndouble *a, int n, int n0) {

  /* petite racine carre */
  if ((n <= n0) || (n <= 2)) {
    if (meth == 0) {
      ndouble *buff = getmem(2*n);
      movemem(buff, a, 2*n);
      sqrtn2(c,a,n);
      freemem(buff);
    }
    else sqrtn2(c,a,n);
  }

  /* grande racine carre */
  else {
    int p=(n-1)/2, q=n-p, r;
    ndouble *buff = getmem(2*n);

    /* racine carre de la partie haute */
    movemem(buff, a, 2*n);
    msqrt(1, c+p, buff+2*p, q, n0);

    /* partie basse par division tronque */
    rm_divh1(c, buff+q-1, c+q-1, p, 2*p+1, n0);
    inc(c, c, n, p);

    /* reste par carr modulaire */
    if (meth) {
      div2(buff, c, n);
      r = remsqrt(a, a, buff, n, 2*n, n, n0);
      while(r) {
	dec(c, &un, n, 1);
	r = 1-inc(a, c, n, n);
	dec(c, &un, n, 1);
      }
    }

    freemem(buff);
  }
}

/* tests */

void sqrt0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  a[2*n-1] = (a[2*n-1] & (mask/4)) | (msb/4);
  movemem(c, a, 2*n);
  msqrt(0, b, c, n, n0);

}

void sqrt1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  a[2*n-1] = (a[2*n-1] & (mask/4)) | (msb/4);
  movemem(c, a, 2*n);
  msqrt(1, b, c, n, n0);

}

int msqrtchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e, *f, r, trois = 3;
  int res;

  e = getmem(n); f = getmem(2*n);
  div2(e, b, n);
  dec(e, &un, n, 1);

  karasqr(f, e, n, n0);
  res = (compare(f, a, 2*n) > 0);
  r = inc(f, b, 2*n, n) + inc(f, b, 2*n, n) + inc(f, &trois, 2*n, 1);
  res |= (!r) && (compare(f, a, 2*n) <= 0);

  freemem(e); freemem(f);
  return(res);
}
