/* +------------------------------------------------------------------------+
   |                                                                        |
   |                        Multiplication de Karatsuba                     |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

/* Mthode nave */
void karatsuba0(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  /* cas de base */
  if (n <= n0) muln2(c,a,b,n);

  /* cas rcursif */
  else {
    int p=n/2, q=n-p;                            /* demi-longueurs */
    int sa,sb;                                   /* signes des diffrences */
    ndouble *buff = getmem(2*q+1);

    sa = asub(c,   a+p, a, q,p);                 /* |a1-a0| */
    sb = asub(c+q, b+p, b, q,p);                 /* |b1-b0| */

    karatsuba0(buff,  c,   c+q, q, n0);          /* |a1-a0||b1-b0| */
    karatsuba0(c,     a,   b,   p, n0);          /* a0b0 */
    karatsuba0(c+2*p, a+p, b+p, q, n0);          /* a1b1 */

    /* buff <- a0b1 + a1b0 - a0b0 */
    if (sa == sb) buff[2*q] = sub(buff, c+2*p, buff, 2*q,2*q) & mask;
    else          buff[2*q] = inc(buff, c+2*p,       2*q,2*q) & mask;

    inc(buff, c,    2*q+1,2*p);                    /* ... + a0b0 */
    inc(c+p,  buff, n+q,2*q+1);                    /* ... = a0b1 + a1b0 */

    freemem(buff);
  }
}

/* Mthode avec recouvrement des additions */
void karatsuba1(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  /* cas de base */
  if (n <= n0) muln2(c,a,b,n);

  /* cas rcursif */
  else {
    int p=n/2, q=n-p;                            /* demi-longueurs */
    ndouble *buff = getmem(2*q);
    int sa,sb;                                   /* signes des diffrences */
    ndouble r1,r2,r3;                              /* retenues des additions */

    sa = asub(c,   a, a+q, q,p);                 /* |a1-a0| */
    sb = asub(c+q, b, b+q, q,p);                 /* |b1-b0| */

    karatsuba1(buff,  c,   c+q, q, n0);          /* |a1-a0||b1-b0| */
    karatsuba1(c,     a,   b,   q, n0);          /* a0b0 */
    karatsuba1(c+2*q, a+q, b+q, p, n0);          /* a1b1 */

    r1 = inc(     c+2*q,   c+q,   q,    q);      /* gamma + beta */
    r2 = add(c+q, c,       c+2*q, q,    q) + r1; /* alpha + beta + gamma */
    r3 = inc(     c+2*q,   c+3*q, q,    2*p-q);  /* beta + gamma + delta */
         inc(     c+2*q,   &r2,   2*p,  1);      /* propage les retenues */
         inc(     c+3*q,   &r1,   2*p-q,1);
         inc(     c+n+p,   &r3,   q,    1);
 
    if (sa==sb) dec(c+q, buff, n+p, 2*q);      /* retranche ...  */
    else        inc(c+q, buff, n+p, 2*q);      /* (a1-a0)(b1-b0) */

    freemem(buff);
  }
}


/* tests */
void mulsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n;
  *sb = n;
  *sc = -2*n;
  *sd = 0;
}

void kara0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  karatsuba0(c, a, b, n, n0);
}

void kara1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  karatsuba1(c, a, b, n, n0);
}

int mulchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;
  int res;

  e = getmem(2*n);
  muln2(e,a,b,n);
  res = compare(c,e,2*n);
  freemem(e);

  return(res);
}
