/* +------------------------------------------------------------------------+
   |                                                                        |
   |              Divisions Rmp-Karp-Markstein et Rmp-directe               |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

static ndouble un = 1;

/* Rmq : il reste des branches de code non testes par manque de donnes
   y faisant passer. Elles sont signales par le commentaire :

                     *** branche non teste ***

   et contiennent des instructions provisoires d'affichage pour dtecter
   un passage par cette branche
*/


                     /* +-------------------------------+
                        |  Division haute avec inverse  |
                        +-------------------------------+ */

/* Entres
     1 <= n < base/2 - 3
     a de longueur p >= n+2
     b de longueur n+1 avec 2b >= base^(n+1)
     c de longueur n+1 tq 0 <= divh(base^(2n+1),b,n-1) - c <= 2

   Sorties
     q <- min(base^n - 1, divh(a*base^(n-1), b, n-1)) de longueur n
     a <- a - high(b,q,n-1)/base^(n-1)

   si on a a < b*base en entre, alors a < base^3 en sortie

   Algorithme
     Si a >= b*base alors q <- base^n - 1,
                          a <- a - high(q,b,n-1)/base^(n-1).

     Sinon on a : 
       a < base^(n+2)
       x = base^(2n+1) - b*c vrifie -(n-1)*base^n < x < 3*base^(n+1)

     Soit u tel que high(a,c,n+1) <= u*base^(n+1) <= a*c obtenu par
     le produit court de Mulders sans limination des retenues venant
     de la partie basse. Donc 0 <= u < b*c/base^n <= base^(n+1)

        q <- floor(u/base)

     Donc 0 <= q < base^n et 0 <= y = a*c - q*base^(n+2) < (n+2)*base^(n+2)

     Soit r = a - high(b,q,n-1)/base^(n-1)
            = a - b*q/base^(n-1) + low(b,q,n-1)/base^(n-1)
            = a - b*(a*c - y)/base^(2n+1) + low(b,q,n-1)/base^(n-1)
            = (a*x + b*y)/base^(2n+1) + low(b,q,n-1)/base^(n-1)

     donc -(n-1)*base < r < (n+5)*base^2 + (n-1)*base.

     Comme n+6 < base/2, r tient sur trois chiffres en complment  2.
     On calcule ces trois chiffres.

     Si r >= b[n]*base:
        s <- floor(r/(b[n-1] + b[n]*base)) < 2*(n+6) < base
        r <- r - (high(b,q+s,n-1) - high(b,q,n-1))/base^(n-1)
        q <- q + s

     maintenant, -(n-2)*(base-1) <= r < b[n-1] + b[n]*base et q <= base^n.

     Si r < 0:
        r <- r + (bigh(b,q,n-1) - high(b,q-1,n-1))/base^(n-1)
        q <- q - 1

     maintenant, 0 <= r < (bigh(b,q+1,n-1) - high(b,q,n-1))/base^(n-1)
     et q < base^n.

        a <- r
	termin

*/

void km_divh_with_inverse(ndouble *q, ndouble *a, ndouble *b, ndouble *c,
                          int n, int p, int n0) {

  ndouble r0,r1,s,t;
  int i;

  /* troncature si a >= b*base */

  while ((p > n+2) && a[p-1] == 0) p--;
  if ((p > n+2) || (compare(a+1,b,n+1) >= 0)) {
    for (i=0; i < n; q[i++] = mask);   /* q <- base^n - 1 */
    dec(a+1, b, p-1, n+1);             /* a <- a - b*base */
    for (i=t=0; i < n; t += b[i++]);
    inc(a,   &t,  p,   1);             /*        + b[0] + .. + b[n-1] */
    inc(a+1, b+n, p-1, 1);             /*        + b[n]*base          */
    return;
  }

  /* q <- floor(u/base) */
  {
    ndouble *buff = getmem(2*n+2);
    highmul(buff, a+1, c, n+1, n0);
    movemem(q, buff+n+1, n);
    freemem(buff);
  }


  /* r0:r1 <- a - high(b*q,n-1)/base^(n-1) */

  r0  = b[n-1]*q[0];
  r1  = b[n]  *q[0] + (r0 >> shift);
  r0 &= mask;
  for (i=1; i < n; i++) {
    r0 += b[n-i-1]*q[i];
    r1 += ((b[n-i+1] << shift) + b[n-i])*q[i] + (r0 >> shift);
    r0 &= mask;
  }

  r0 = a[0] - r0;
  r1 = (a[2] << shift) + a[1] - r1;
  if (r0 >> shift) {r0 &= mask; r1--;}

  /* si r0:r1 >= b[n]*base, divise par b[n]:b[n-1] et corrige q,r0:r1  */
  /* on ne propage pas la retenue de rang n s'il y en a une, car alors */
  /* q = base^n et on va le dcrmenter juste aprs.                   */

  if ((zdouble)r1 >= b[n]) {
    s   = r1/b[n];
    t   = s*b[n-1];
    r0 -= t & mask; r1 -= (t >> shift) + s*b[n];
    if (r0 >> shift) {r0 &= mask; r1--;}
    while ((zdouble)r1 < 0) {
      s--;
      r0 += b[n-1]; r1 += b[n] + (r0 >> shift); r0 &= mask;
    }

    for (i=0; (s) && (i < n); i++) {
      s   += q[i];
      q[i] = s & mask;
      s  >>= shift;
      if ((s) && (i < n-1)) {
        r0 -= b[n-2-i]; if (r0 >> shift) {r0 &= mask; r1--;}
      }
    }
  }

  /* si r0:r1 < 0, diminue q */

  if ((zdouble)r1 < 0) {
    /* *** branche non teste *** */
    r0 += b[n-1];
    r1 += b[n] + (r0 >> shift);
    r0 &= mask;
    for (i=0; (i < n) && (q[i] == 0); i++) {
      q[i] = mask;
      if (i < n-1) {r0 += b[n-2-i]; r1 += (r0 >> shift); r0 &= mask;}
    }
    if (i < n) q[i]--;
    fprintf(stderr,"reste ngatif dans km_divh_with_inverse, i=%d\n",i);
  }

  /* termin */
  a[0] = r0; a[1] = r1 & mask; a[2] = r1 >> shift;
  clearmem(a+3,p-3);

}


                              /* +-------------+
                                 |  Inversion  |
                                 +-------------+ */

/* Entres
     1 <= n <= base - 6
     b de longueur n avec 2b >= base^n

   Sorties
     a <- divh(base^(2n-1), b, n-2) de longueur n
     r <- (base^(2n-1) - high(a, b, n-2))/base^(n-2) de longueur 3

   Algorithme

     Si n <= max(n0,2) alors diviser navement.
     Sinon :
       p <- ceil(n/2)-1, q <- floor(n/2)+1 (q < n pour n > 2)
       soient a = a0 + base^p*a1, b = b0 + base^p*b1 = b0' + base^(q-1)*b1'

       a1 <- divh(base^(2q-1), b1, q-2)
       r  <- (base^(2q-1) - high(a1, b1, q-2))/base^(q-2)

       s  <- (base^(p+2q-1) - high(a1, b, q-2))/base^(q-2)
           = r*base^p - med(a1, b, q-2, n-2)/base^(q-2)
	              - a1[q-1]*b[p-1]*base^p

     donc a[q-1] <= 2 et -(q+2)*base^(p+1) < s < base^(p+3).

     Si s < 0 :
       s  <- s + (high(a1, b, q-2) - high(a1-1, b, q-2))/base^(q-2)
       a1 <- a1-1

     Maintenant, 0 <= s < base^(p+3).

       a0 <- divh(s*base^(n-2), b1', n-2)

    a0 est calcul par km_divh_with_inverse  l'aide de a1 (cas n impair)
    ou floor(a1/base) (cas n pair) comme inverse approch de b'1. Dans les
    deux cas, l'inverse approch diffre de divh(base^(2q-1), b1', p-1)
    d'au plus deux units (une dcrmentation dans la procdure en cours
    et une dcrmentation ventuelle dans l'appel rcursif si n est pair).

*/
   

void km_inv(ndouble *a, ndouble *r, ndouble *b, int n, int n0) {

  /* cas de base */
  if (n == 1) {a[0] = 1; r[0] = 0; r[1] = (-b[0]) & mask; r[2] = 0;}
  else if ((n <= n0) || (n <= 2)) {
    ndouble *buff = getmem(2*n+3);
    movemem(buff+n+3, b, n);
    clearmem(buff,n+3); buff[n+1] = 1;
    divn2h(a, buff, buff+n+2, n, n+2);
    movemem(r, buff, 3);
    freemem(buff);
  }

  /* cas rcursif */
  else {
    int q = n/2+1, p = n-q;
    ndouble *s, t;
    zdouble z;
    int i;

    /* inverse la partie haute */
    km_inv(a+p, r, b+p, q, n0);

    /* s <- med(a1, b, q-2)/base^(q-2) + a1[q-1]*b[p-1]*base^p */
    s = getmem(p+3);
    if (p == q-2) {

      /* n pair : produit mdian (2p-1)*p + deux colonnes */
      mpmul(s, b+1, a+p, p, n0);
      for (i=t=0; i < p; i++) {
        t += s[i] + a[n-2]*b[i];
        s[i] = t & mask;
        t >>= shift;
      }
      s[p+2] = inc(s+p, &t, 2, 1);
      for (i=t=0; i < p; i++) {
        t += s[i+1] + a[n-1]*b[i];
        s[i+1] = t & mask;
        t >>= shift;
      }
      inc(s+p+1, &t, 2, 1);
    }

    else {
      /* n impair : produit mdian (2p-1)*p + une colonne */
      mpmul(s, b, a+p, p, n0);
      for (i=t=0; i < p; i++) {
        t += s[i+1] + a[n-1]*b[i];
        s[i+1] = t & mask;
        t >>= shift;
      }
      t += s[p+1];
      s[p+1] = t & mask;
      s[p+2] = t >> shift;
    }

    /* s <- r*base^p - s */
    z  = neg(s,s,p+3);
    z += inc(s+p, r, 3, 3);

    /* si z < 0 alors a est trop grand, corrige */
    if (z) {
      /* *** branche partiellement teste *** */
      inc(s, b+q-2, p+3, p+2);
      for (i=p; a[i] == 0; i++) {
        a[i] = mask;
        if (i < n-2) inc(s, b+n-3-i, p+3, 1);
      }
      a[i]--;
    if (i > p) fprintf(stderr,"reste ngatif dans km_inv, i-p=%d\n",i-p);
    }

    /* termine avec km_divh_with_inverse */
    km_divh_with_inverse(a, s, b+q-1, a+q-1, p, p+3, n0);
    movemem(r, s, 3);
    freemem(s);
  }
}


                  /* +-------------------------------------+
                     |  Division haute par Karp-Markstein  |
                     +-------------------------------------+ */


/* Entres
     1 <= n <= base-7
     b de longueur n+1 avec 2b >= base^(n+1)
     a de longueur n+2 avec a < b*base

   Sorties
     c <- divh(a*base^(n-1), b, n-1) de longueur n
     a <- a - high(b,c,n-1)/base^(n-1) < base^3

   Algorithme

     Si n <= n0 alors diviser navement
     Sinon :
       p <- floor(n/2), q <- ceil(n/2)
       soit a = a0 + base^p*a1,
            b = b0 + base^p*b1 = b'0 + base^q*b1',
            c = c0 + base^p*c1

       d  <- divh(base^(2q+1), b1, q-1) calcul par km_inv

       c1 <- min(base^q - 1, divh(a1*base^(q-1), b1, q-1))
       r  <- a1 - high(b1, c1, q-1)/base^(q-1)
       calculs par km_divh_with_inverse avec d comme inverse approch

       a  <- a - high(b, c1, q-1)/base^(q-1)
           = a0 + base^p*r - med(b,c1,q-1,n-1)/base^(q-1)

       donc -q*base^(p+1) < a < base^(p+3)
       Si a < 0:
          a  <- a + (high(b, c1, q-1) - high(b, c1-1, q-1))/base^(q-1)
	  c1 <- c1-1

       maintenant 0 <= a < base^(p+3)

       c0 <- min(base^p - 1, divh(a*base^(p-1), b1', p-1))
       a  <- a - high(b1', c0, p-1)/base^(p-1)
       calculs par km_divh_with_inverse avec d (cas n pair) ou
       floor(d/base) (cas n impair) comme inverse approch

*/

void km_divh(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  /* division nave */
  if (n <= n0) divn2h(c, a, b, n, n+2);

  /* division par inversion */
  else {
    int p = n/2, q = n-p;
    ndouble *buff = getmem(n+3), *d = buff+p+2, r[3];
    ndouble s;
    zdouble z;
    int i;

    /* inverse de la partie haute de b */
    km_inv(d, r, b+p, q+1, n0);

    /* partie haute du quotient */
    km_divh_with_inverse(c+p, a+p, b+p, d, q, q+2, n0);

    /* retranche med(b,c1) */
    mpmul(buff, b, c+q, p, n0);
    if (p < q) {
      for (i=s=0; i < p; i++) {
	s += buff[i] + b[i+p]*c[p];
	buff[i] = s & mask;
	s >>= shift;
      }
      s += buff[p]; buff[p] = s & mask;
      buff[p+1] += s >> shift;
    }
    z = dec(a, buff, p+3, p+2);

    /* si < 0, correction */
    if (z) {
      /* *** branche partiellement teste *** */
      inc(a, b+q-1, p+3, p+2);
      for (i=p; c[i] == 0; i++) {
	c[i] = mask;
	inc(a, b+n-1-i, p+3, 1);
      }
      c[i]--;
    if (i > p) fprintf(stderr,"reste ngatif dans km_divh, i-p=%d\n",i-p);
    }

    /* partie basse */
    km_divh_with_inverse(c, a, b+q, d+q-p, p, p+3, n0);

    /* termin */
    freemem(buff);

  }
}

                  /* +---------------------------+
                     |  Division haute avec Rmp  |
                     +---------------------------+ */

/* Entres
     1 <= n <= base-7
     b de longueur n+1 avec 2b >= base^(n+1)
     a de longueur p >= n+2

   Sorties
     c <- divh(a*base^(n-1), b, n-1) de longueur n
     a <- a - high(b,c,n-1)/base^(n-1)

   si on a a < b*base en entre, alors a < base^3 en sortie

   Algorithme

     Si a >= b*base alors c <- base^n - 1,
                          a <- a - high(c,b,n-1)/base^(n-1).

     Si n <= n0 alors diviser navement
     Sinon :
       p <- floor(n/2), q <- ceil(n/2)
       soit a = a0 + base^p*a1,
            b = b0 + base^p*b1 = b'0 + base^q*b1',
            c = c0 + base^p*c1

       c1 <- min(base^q - 1, divh(a1*base^(q-1), b1, q-1))
       r  <- a1 - high(b1, c1, q-1)/base^(q-1)
       calculs rcursivement

       a  <- a - high(b, c1, q-1)/base^(q-1)
           = a0 + base^p*r - med(b,c1,q-1,n-1)/base^(q-1)

       donc -q*base^(p+1) < a < base^(p+3)
       Si a < 0:
          a  <- a + (high(b, c1, q-1) - high(b, c1-1, q-1))/base^(q-1)
	  c1 <- c1-1

       maintenant 0 <= a < base^(p+3)

       c0 <- min(base^p - 1, divh(a*base^(p-1), b1', p-1))
       a  <- a - high(b1', c0, p-1)/base^(p-1)
       calculs rcursivement

*/

void rm_divh(ndouble *c, ndouble *a, ndouble *b, int n, int p, int n0) {

  /* troncature si a >= b*base */
  while ((p > n+2) && a[p-1] == 0) p--;
  if ((p > n+2) || (compare(a+1,b,n+1) >= 0)) {
    int i;
    ndouble t;

    for (i=0; i < n; c[i++] = mask);   /* c <- base^n - 1 */
    dec(a+1, b, p-1, n+1);             /* a <- a - b*base */
    for (i=t=0; i < n; t += b[i++]);
    inc(a,   &t,  p,   1);             /*        + b[0] + .. + b[n-1] */
    inc(a+1, b+n, p-1, 1);             /*        + b[n]*base          */
    return;
  }

  /* division nave */
  if (n <= n0) divn2h(c, a, b, n, n+2);

  /* division rcursive */
  else {
    int p = n/2, q = n-p;
    ndouble *buff;
    ndouble s;
    zdouble z;
    int i;

    /* partie haute du quotient */
    rm_divh(c+p, a+p, b+p, q, q+2, n0);

    /* retranche med(b,c1) */
    buff = getmem(p+2);
    mpmul(buff, b, c+q, p, n0);
    if (p < q) {
      for (i=s=0; i < p; i++) {
	s += buff[i] + b[i+p]*c[p];
	buff[i] = s & mask;
	s >>= shift;
      }
      s += buff[p]; buff[p] = s & mask;
      buff[p+1] += s >> shift;
    }
    z = dec(a, buff, p+3, p+2);
    freemem(buff);

    /* si < 0, correction */
    if (z) {
      /* *** branche partiellement teste *** */
      inc(a, b+q-1, p+3, p+2);
      for (i=p; c[i] == 0; i++) {
	c[i] = mask;
	inc(a, b+n-1-i, p+3, 1);
      }
      c[i]--;
    if (i > p) fprintf(stderr,"reste ngatif dans rm_divh, i-p=%d\n",i-p);
    }

    /* partie basse */
    rm_divh(c, a, b+q, p, p+3, n0);

  }
}

/* idem rm_divh, mais le reste intermdiaire est calcul par produit modulaire */
void rm_divh1(ndouble *c, ndouble *a, ndouble *b, int n, int p, int n0) {

  /* troncature si a >= b*base^n */
  while ((p > 2*n+1) && a[p-1] == 0) p--;
  if ((p > 2*n+1) || (compare(a+n,b,n+1) >= 0)) {
    int i;
    for (i=0; i < n; c[i++] = mask);   /* c <- base^n - 1   */
    return;
  }

  /* rm_divh est plus rapide au dessous de ce seuil */
  if (n <= 16*n0) rm_divh(c, a+n-1, b, n, n+2, n0);

  /* grande division */
  else {
    int p = n/2, q = n-p, r;
    ndouble *buff;

    /* partie haute du quotient */
    buff = getmem(2*q+1);
    movemem(buff, a+n+p-q, 2*q+1);
    rm_divh1(c+p, buff, b+p, q, 2*q+1, n0);
    freemem(buff);

    /* a <- a - b*c1*base^p */
    r = remdiv(a+p, a+p, b, c+p, n+2, n+q+1, n+1, q, n0);
    if (r) {dec(c+p,&un, q, 1); inc(a+p, b, n+2, n+1);}

    /* partie basse */
    rm_divh1(c, a+q, b+q, p, 2*p+2, n0);

  }
}


                     /* +------------------------------+
                        |  teste les divisions hautes  |
                        +------------------------------+ */


void kmdsz0(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n+1;
  *sb = -(n+1);
  *sc = 0;
  *sd = 0;
}

void kmdsz1(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n+2;
  *sb = n+1;
  *sc = -(n+1);
  *sd = 0;
}

void kmd0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble r[3];

  a[n] |= msb; a[0] = b[0] = 0;
  km_inv(b+1, r, a+1, n, n0);
}

void kmd1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;

  b[n] |= msb; a[n+1] %= b[n]; c[0] = 0;
  e = getmem(n+2); movemem(e, a, n+2);
  km_divh(c+1, e, b, n, n0);
  freemem(e);
}

void rdiv0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;

  b[n] |= msb; a[n+1] %= b[n]; c[0] = 0;
  e = getmem(n+2); movemem(e, a, n+2);
  rm_divh(c+1, e, b, n, n+2, n0);
  freemem(e);
}

int divhchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int res;
  ndouble *y, *z, r;

  y = getmem(n+2); z = getmem(n+1);

  /* vrifie que high(c,b,n-1) <= a */
  shortmul_h(y, c, b, n+1, n0);
  res = (compare(y, a, n+2) > 0);

  /* vrifie que high(c+1,b,n-1) > a */
  z[0] = 0; r = add(z+1, c+1, &un, n, 1);
  shortmul_h(y, z, b, n+1, n0);
  res |= (!r) && (compare(a, y, n+2) >= 0);

  freemem(y); freemem(z);
  return(res);

}

int invhchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  int res;
  ndouble *x;

  x = getmem(n+2); clearmem(x, n+1); x[n+1] = 1;
  res = divhchk(x, a, b, d, n, n0);
  freemem(x);
  return(res);

}

void rdivsz2(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = 2*n+1;
  *sb = n+1;
  *sc = -(n+1);
  *sd = 0;
}

void rdiv2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e = getmem(2*n+1);

  b[n] |= msb; a[2*n] %= b[n]; c[0] = 0;
  movemem(e, a, 2*n+1);
  rm_divh1(c+1, e, b, n, 2*n+1, n0);
  freemem(e);
}

int mdivhchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble r, *y, *z;
  zdouble s;
  int res;

  y = getmem(2*n+2); z = getmem(n+1);

  /* vrifie que b*(c-1) <= a */
  z[0] = 0; s = sub(z+1, c+1, &un, n, 1);
  karamul(y, z, b, n+1, n0);
  res = (!s) && (compare(y+1, a, 2*n+1) > 0);

  /* vrifie que b*(c+1) > a */
  r = add(z+1, c+1, &un, n, 1);
  karamul(y,z, b, n+1, n0);
  res |= (!r) && (compare(y+1, a, 2*n+1) <= 0);

  freemem(y); freemem(z);
  return(res);
}


                          /* +---------------------+
                             |  Division complte  |
                             +---------------------+ */

/* Entres
    1 <= n <= base-7
    b de longueur n avec 2b >= base^n
    a de longueur 2n avec a < b*base^n
    meth = 0 -> Karp-Markstein avec reste
	   1 -> Rmp direct avec reste

  Sorties
    q <- floor(a/b) de longueur n
    a <- a - b*q

  Algorithme

  Si n <= max(n0,2) diviser navement
  Sinon :

    q <- divh(floor(a/base^(n-2), b*base, n-1)) calcul par km_divh ou rm_divh
    a <- a - b*q mod ppcm(base^(n/4+1) - 1, .., B^(n/4+4) - 1)
    si a >= base^n :
      a <- a + b mod base^n
      q <- q - 1
*/

void kmr_div(int meth, ndouble *q, ndouble *a, ndouble *b, int n, int n0) {

  /* petite division */
  if ((n <= n0) || (n <= 2)) divn2(q, a, b, n);

  /* grande division */
  else {
    ndouble *buff = getmem(2*n+3);
    int r;

    buff[0] = 0; movemem(buff+1,b,n); movemem(buff+n+1, a+n-2, n+2);
    switch(meth) {
    case 0: km_divh(q, buff+n+1, buff, n,      n0); break;
    case 1: rm_divh(q, buff+n+1, buff, n, n+2, n0); break;
    }
    freemem(buff);

    r = remdiv(a, a, b, q, n, 2*n, n, n, n0);
    if (r) {dec(q, &un, n, 1); inc(a, b, n, n);}

  }
}

/* idem avec rm_divh1 */
void kmr_div1(ndouble *q, ndouble *a, ndouble *b, int n, int n0) {

  /* petite division */
  if ((n <= n0) || (n <= 2)) divn2(q, a, b, n);

  /* grande division */
  else {
    ndouble *buff = getmem(3*n+6);
    int r;

    buff[0]   = 0; movemem(buff+1,b,n);
    buff[n+1] = 0; movemem(buff+n+2, a, 2*n);
    rm_divh1(q, buff+n+1, buff, n, n+2, n0);
    freemem(buff);

    r = remdiv(a, a, b, q, n, 2*n, n, n, n0);
    if (r) {dec(q, &un, n, 1); inc(a, b, n, n);}

  }
}

              /* +----------------------------------+
                 |  Teste les divisions avec reste  |
                 +----------------------------------+ */

void kmd2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  b[n-1] |= msb; a[2*n-1] %= b[n-1];
  movemem(d, a, 2*n);
  kmr_div(0, c, d, b, n,n0);

}

void rdiv1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  b[n-1] |= msb; a[2*n-1] %= b[n-1];
  movemem(d, a, 2*n);
  kmr_div(1, c, d, b, n,n0);

}

void rdiv3(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {

  b[n-1] |= msb; a[2*n-1] %= b[n-1];
  movemem(d, a, 2*n);
  kmr_div1(c, d, b, n,n0);

}
