/* +------------------------------------------------------------------------+
   |                                                                        |
   |                        Produit tronqu de Mulders                      |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

/* moiti basse */
void mulders_low(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {
  int p = 0.6942363715*n, q=n-p;

  if ((n <= n0) || (q >= p)) muln2l(c,a,b,n);
  else {
    int i;
    ndouble *buff = getmem(2*p);
    ndouble r;

    /* produit central, retire les produits dpassant la diagonale */
    karamul(buff,a,b,p,n0);
    movemem(c,buff,n);
    r = buff[n];
    if (n+1 < 2*p) r += buff[n+1] << shift;
    for (i=q+1; i<p; i++) {
      r -= a[i]*b[n-i];
      if (i>q+1) r -= (a[i]*b[n+1-i]) << shift;
    }
    c[n]   = r & mask;
    c[n+1] = (r >> shift) & mask;

    /* triangles latraux */
    if (q > 0){
      mulders_low(buff, a, b+p, q, n0); inc(c+p, buff, q+2, q+2);
      mulders_low(buff, a+p, b, q, n0); inc(c+p, buff, q+2, q+2);
    }

    freemem(buff);
  }
}

/* moiti basse, on coupe  la moiti */
void mul_low(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {

  /* cas de base */
  if (n <= n0) muln2l(c,a,b,n);

  /* cas rcursif */
  else {
    int p=n/2, q=n-p;
    ndouble *buff = getmem(p+2);

    karamul(c,a,b,q,n0);      clearmem(c+2*q,n+2-2*q);
    mul_low(buff,a,b+q,p,n0); inc(c+q,buff,n+2-q,p+2);
    mul_low(buff,a+q,b,p,n0); inc(c+q,buff,n+2-q,p+2);

    freemem(buff);
  }
}

/* moiti haute */
void mulders_high(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {
  int p = 0.6942363715*n, q=n-p;

  if ((n <= n0) || (q >= p)) muln2h(c,a,b,n);
  else {
    int i,j;
    ndouble *buff = getmem(2*p), t[3];
    ndouble r;

    /* produit central, retire les produits sous la diagonale */
    karamul(buff, a+q, b+q, p,n0);
    clearmem(t,3);
    for (i=q,j=n-q-2; j>=q; i++,j--) {
      r = a[i]*b[j] + t[0];
      if (j > q) r += (a[i]*b[j-1]) >> shift;
      t[0] = r & mask; r = (r >> shift) + t[1];
      t[1] = r & mask; r = (r >> shift) + t[2];
      t[2] = r & mask;
    }
    dec(buff+p-q-2,t,n+2,3);
    movemem(c,buff+p-q-1,n+1);

    /* triangles latraux */
    if (q > 0){
      mulders_high(buff, a, b+p, q, n0); inc(c, buff, n+1, q+1);
      mulders_high(buff, a+p, b, q, n0); inc(c, buff, n+1, q+1);
    }

    freemem(buff);
  }
}

/* produit tronqu, partie haute, sans limination des produits infrieurs */
void raw_mulders_high(ndouble *c, ndouble *a, ndouble *b, int n, int n0) {
  int p = 0.6942363715*n, q=n-p;

  if ((n <= n0) || (q >= p)) muln2h(c+n-1,a,b,n);
  else {

    /* produit central */
    karamul(c+2*q, a+q, b+q, p,n0);

    /* triangles latraux */
    if (q > 0){
      ndouble *buff = getmem(2*q);
      raw_mulders_high(buff, a, b+p, q, n0); inc(c+n-1, buff+q-1, n+1, q+1);
      raw_mulders_high(buff, a+p, b, q, n0); inc(c+n-1, buff+q-1, n+1, q+1);
      freemem(buff);
    }
  }
}

/* tests */
void mullsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n;
  *sb = n;
  *sc = -(n+2);
  *sd = 0;
}

void mulhsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n;
  *sb = n;
  *sc = -(n+1);
  *sd = 0;
}

void mul0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  mulders_low(c, a, b, n, n0);
}

void mul1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  mulders_high(c, a, b, n, n0);
}

void mul2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  mul_low(c, a, b, n, n0);
}

int mullchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;
  int res;

  e = getmem(n+2);
  muln2l(e,a,b,n);
  res = compare(c,e,n+2);
  freemem(e);

  return(res);
}

int mulhchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;
  int res;

  e = getmem(n+1);
  muln2h(e,a,b,n);
  res = compare(c,e,n+1);
  freemem(e);

  return(res);
}
