/* +------------------------------------------------------------------------+
   |                                                                        |
   |                           lvation au carr                           |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"

/* c <- a^2, mthode de Karatsuba */
void karasqr0(ndouble *c, ndouble *a, int n, int n0) {

  /* cas de base */
  if (n <= n0) sqrn2(c,a,n);

  /* cas rcursif */
  else {
    int p=n/2, q=n-p;                            /* demi-longueurs */
    ndouble *buff = getmem(2*q+1);

    asub(c,   a+p, a, q,p);                      /* |a1-a0| */

    karasqr0(buff,  c,   q, n0);                /* (a1-a0)^2 */
    karasqr0(c,     a,   p, n0);                /* a0^2      */
    karasqr0(c+2*p, a+p, q, n0);                /* a1^2      */

    /* buff <- 2a0a1 - a0^2 */
    buff[2*q] = sub(buff, c+2*p, buff, 2*q,2*q) & mask;

    inc(buff, c,    2*q+1,2*p);                    /* ... + a0^2  */
    inc(c+p,  buff, n+p,2*q+1);                    /* ... = 2a0a1 */

    freemem(buff);
  }
}

/* c <- low(a^2) par shortprod */
void shortsqr0(ndouble *c, ndouble *a, int n, int n0) {

  /* cas de base */
  if (n <= n0) sqrn2l(c,a,n);

  /* cas rcursif */
  else {
    int p=n/2, q=n-p, i;
    ndouble *buff = getmem(p+2);
    ndouble r;

    karasqr0(c,a,q,n0);              /* a0^2 */
    shortmul_l(buff,a,a+q,p,n0);     /* low(a0*a1) */

    clearmem(c+2*q,n+2-2*q);
    for (i=r=0; i<p+2; i++) {
      r += 2*buff[i] + c[i+q];
      c[i+q] = r & mask;
      r >>= shift;
    }

    freemem(buff);
  }
}

/* c <- low(a^2) par rmp */
void shortsqr1(ndouble *c, ndouble *a, int n, int n0) {

  /* cas de base */
  if ((n <= n0) || (n <= 3)) sqrn2l(c,a,n);

  /* cas rcursif */
  else {
    int p=n/2, q=n-p, i;
    ndouble *buff = getmem(p+2);
    ndouble r,r0,r1;

    /* buff <- mp([a0,2a1],a0) */
    movemem(c,a+q-p+1,2*p-1-q);
    for (i=p, r=r0=r1=0; i < n; i++) {
      r += 2*a[i];
      c[p-q-1+i] = r & mask;
      r >>= shift;
      if (r) {
        if (i<q) r0 += a[q-1-i];
        else     r1 += a[n-1-i];
      }
    }
    mpmul(buff, c, a, p, n0);
    dec(buff,   &r0, p+2, 1);
    inc(buff+p, &r1,   2,1);

    /* carr manquant si n est impair */
    if (p<q) {r = a[p]*a[p]; inc(buff+p-1, &r, 3,1);}

    /* c <- low(a0^2) */
    shortsqr1(c,a,q,n0);

    /* assemblage */
    movemem(c+q+2, buff+2, p);
    inc(c+q,buff,p+2,2);

    freemem(buff);
  }
}

/* tests */
void sqrsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n;
  *sb = -2*n;
  *sc = 0;
  *sd = 0;
}

void sqr0(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  karasqr0(b, a, n, n0);
}

int sqrchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;
  int res;

  e = getmem(2*n);
  sqrn2(e,a,n);
  res = compare(b,e,2*n);
  freemem(e);

  return(res);
}

void sqrlsz(int n, int *sa, int *sb, int *sc, int *sd) {
  *sa = n;
  *sb = -(n+2);
  *sc = 0;
  *sd = 0;
}

void sqr1(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  shortsqr0(b, a, n, n0);
}

void sqr2(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  shortsqr1(b, a, n, n0);
}

int sqrlchk(ndouble *a, ndouble *b, ndouble *c, ndouble *d, int n, int n0) {
  ndouble *e;
  int res;

  e = getmem(n+2);
  sqrn2l(e,a,n);
  res = compare(b,e,n+2);
  freemem(e);

  return(res);
}
