/* +------------------------------------------------------------------------+
   |                                                                        |
   |                Chronomtrage des algorithmes multiprcision            |
   |                                                                        |
   +------------------------------------------------------------------------+ */

#include "defs.h"
#include <time.h>


                   /* +-------------------------+
                      |  description des tests  |
                      +-------------------------+ */

typedef void (*fsize)(int, int *, int *, int *, int *);
typedef void (*fcalc)(ndouble *, ndouble *, ndouble *, ndouble *, int, int);
typedef int  (*fctrl)(ndouble *, ndouble *, ndouble *, ndouble *, int, int);

typedef struct {
  char  *name;    /* nom symbolique           */
  char  *desc;    /* description              */
  fcalc calc;     /* fonction  appeler       */
  fsize size;     /* taille des paramtres    */
  fctrl chk;      /* fonction de contrle     */
  int   sel;      /* algorithme slectionn ? */
} t_test;

/*
 La fonction [calc] effectue le calcul avec 4 paramtres de type ndouble *
 qui peuvent tre des arguments ou des rsultats, la taille n slectionne
 et le seuil de Karatsuba, n0.

 La fonction [size] dfinit les tailles des quatre paramtres en fonction de
 n. Une taille nulle indique que le paramtre correspondant n'est pas
 utilis ; une taille ngative indique qu'il s'agit d'un rsultat, la taille
 rellement ncessaire tant la valeur absolue du nombre retourn.

 La fonction [chk] examine les paramtres aprs calcul et retourne 0 si les
 rsultats sont corrects, 1 s'il y a une erreur.

 Rmq: [calc] peut modifier les arguments transmis, par ex. pour imposer un
 bit de poids fort valant 1, sous rserve que :

 1. [calc] pouvant tre appele plusieurs fois avec les mmes paramtres, elle
    doit effectuer les mmes calculs  chaque appel (i.e. les modifs doivent
    tre idempotentes et avoir lieu avant le dbut effectif du calcul) ;

 2. [chk] doit pouvoir dterminer si le calcul est juste  partir des
    paramtres retourns par [calc].

*/

t_test test[] = {

  {"kara0", "Multiplication de Karatsuba ordinaire",     kara0,  mulsz,   mulchk,    0},
  {"kara1", "Multiplication de Karatsuba conomique",    kara1,  mulsz,   mulchk,    0},

  {"rmp0",  "Produit mdian avec retenues",              mp0,    mpsz,    mpchk,     0},
  {"rmp1",  "Produit mdian par blocs",                  mp1,    mpsz,    mpchk,     0},

  {"mul0",  "Produit court de Mulders, partie basse",    mul0,   mullsz,  mullchk,   0},
  {"mul1",  "Produit court de Mulders, partie haute",    mul1,   mulhsz,  mulhchk,   0},
  {"mul2",  "Produit court naIf, partie basse",          mul2,   mullsz,  mullchk,   0},

  {"mmul0", "Produit modulo B^n-1",                      mmul0,  mmulsz0, mmulchk0 , 0},
  {"mmul1", "Produit modulo B^(n/2)-1, B^(n/2+1)-1",     mmul1,  mmulsz1, mmulchk1,  0},
  {"mmul2", "Produit modulo B^(n/3)-1 .. B^(n/3+2)-1",   mmul2,  mmulsz2, mmulchk2,  0},
  {"mmul3", "Produit modulo B^(n/4)-1 .. B^(n/4+3)-1",   mmul3,  mmulsz3, mmulchk3,  0},
  {"mmul4", "Produit modulo B^(2n/3)-1 .. B^(2n/3+2)-1", mmul4,  mmulsz4, mulchk,    0},
  {"mmul5", "Produit modulo B^(2n/4)-1 .. B^(2n/4+3)-1", mmul5,  mmulsz5, mulchk,    0},

  {"sqr0",  "Carr de Karatsuba",                        sqr0,   sqrsz,   sqrchk,    0},
  {"sqr1",  "Carr court de Mulders, partie basse",      sqr1,   sqrlsz,  sqrlchk,   0},
  {"sqr2",  "Carr court RMP, partie basse",             sqr2,   sqrlsz,  sqrlchk,   0},

  {"msqr0", "Carr modulo B^n-1",                        msqr0,  msqrsz0, msqrchk0,  0},
  {"msqr1", "Carr modulo B^(n/2)-1, B^(n/2+1)-1",       msqr1,  msqrsz1, msqrchk1,  0},
  {"msqr2", "Carr modulo B^(n/3)-1 .. B^(n/3+2)-1",     msqr2,  msqrsz2, msqrchk2,  0},
  {"msqr3", "Carr modulo B^(n/4)-1 .. B^(n/4+3)-1",     msqr3,  msqrsz3, msqrchk3,  0},
  {"msqr4", "Carr modulo B^(2n/3)-1 .. B^(2n/3+2)-1",   msqr4,  msqrsz4, sqrchk,    0},
  {"msqr5", "Carr modulo B^(2n/4)-1 .. B^(2n/4+3)-1",   msqr5,  msqrsz5, sqrchk,    0},

  {"burn0", "Division de Burnikel-Ziegler sans reste",   burn0,  burnsz,  divnrchk,  0},
  {"burn1", "Division de Burnikel-Ziegler avec reste",   burn1,  burnsz,  divchk,    0},

  {"kmd0",  "Inversion de Karp-Markstein",               kmd0,   kmdsz0,  invhchk,   0},
  {"kmd1",  "Division de Karp-Markstein sans reste",     kmd1,   kmdsz1,  divhchk,   0},
  {"kmd2",  "Division de Karp-Markstein avec reste",     kmd2,   burnsz,  divchk,    0},

  {"rdiv0", "Division par RMP sans reste",               rdiv0,  kmdsz1,  divhchk,   0},
  {"rdiv1", "Division par RMP avec reste",               rdiv1,  burnsz,  divchk,    0},
  {"rdiv2", "Division modulaire sans reste",             rdiv2,  rdivsz2, mdivhchk,  0},
  {"rdiv3", "Division modulaire avec reste",             rdiv3,  burnsz,  divchk,    0},

  {"zimm0", "Racine carre de Zimmermann sans reste",    zimm0,  sqrtsz,  sqrtnrchk, 0},
  {"zimm1", "Racine carre de Zimmermann avec reste",    zimm1,  sqrtsz,  sqrtchk,   0},

  {"sqrt0", "Racine carre modulaire sans reste",        sqrt0,  sqrtsz,  msqrtchk,  0},
  {"sqrt1", "Racine carre modulaire avec reste",        sqrt1,  sqrtsz,  sqrtchk,   0}

};

#define nbtests (sizeof(test)/sizeof(t_test))


                     /* +--------------------+
                        |  Effectue un test  |
                        +--------------------+ */

double do_test(t_test *tst, int n, int n0, int time, int check) {
  int sa, sb, sc, sd, nrep;
  double t;

  /* taille des oprandes */
  tst->size(n, &sa, &sb, &sc, &sd);  

  /* Compte le nb d'itrations pour atteindre le temps demand.     */
  /* On effectue en fait une rgle de 3  partir du nb d'itrations */
  /* constat pour une dure de 100ms                               */
  {
    ndouble *a, *b, *c, *d;
    
    a = (sa) ? getmem(abs(sa)) : NULL; if (sa > 0) random_init(a, sa);
    b = (sb) ? getmem(abs(sb)) : NULL; if (sb > 0) random_init(b, sb);
    c = (sc) ? getmem(abs(sc)) : NULL; if (sc > 0) random_init(c, sc);
    d = (sd) ? getmem(abs(sd)) : NULL; if (sd > 0) random_init(d, sd);

    t = chrono();
    nrep = 0;
    do {
      tst->calc(a,b,c,d,n,n0);
      nrep++;
    } while (chrono()-t < 100);
    nrep = (nrep*time + 99)/100;


    if (sa) freemem(a);
    if (sb) freemem(b);
    if (sc) freemem(c);
    if (sd) freemem(d);
  }

  /* contrle */
  if (check) {
    ndouble *a, *b, *c, *d;
    
    a = (sa) ? getmem(abs(sa)) : NULL;
    b = (sb) ? getmem(abs(sb)) : NULL;
    c = (sc) ? getmem(abs(sc)) : NULL;
    d = (sd) ? getmem(abs(sd)) : NULL;

    for ( ; nrep; nrep--) {

      if (sa > 0) random_init(a, sa);
      if (sb > 0) random_init(b, sb);
      if (sc > 0) random_init(c, sc);
      if (sd > 0) random_init(d, sd);

      tst->calc(a,b,c,d,n,n0);
      if (tst->chk(a,b,c,d,n,n0)) {
	fprintf(stderr,"erreur dans %s n=%d n0=%d\n", tst->name, n, n0);
	if (sa) {fprintf(stderr,"a="); dump(a,abs(sa));}
	if (sb) {fprintf(stderr,"b="); dump(b,abs(sb));}
	if (sc) {fprintf(stderr,"c="); dump(c,abs(sc));}
	if (sd) {fprintf(stderr,"d="); dump(d,abs(sd));}
	fflush(stderr);
	exit(1);
      }

    }

    if (sa) freemem(a);
    if (sb) freemem(b);
    if (sc) freemem(c);
    if (sd) freemem(d);

    return(0);
  }

  /* chronomtrage : on mmorise les arguments des [nrep] appels avant */
  /* de dmarrer le chronomtre de faon  ne pas faire intervenir le  */
  /* gnrateur alatoire dans les temps mesurs. Les rsultats sont   */
  /* communs  tous les appels.                                        */
  else {
    ndouble **a, **b, **c, **d;
    int i;

    a = (ndouble **)malloc(nrep*sizeof(ndouble *));
    b = (ndouble **)malloc(nrep*sizeof(ndouble *));
    c = (ndouble **)malloc(nrep*sizeof(ndouble *));
    d = (ndouble **)malloc(nrep*sizeof(ndouble *));

    if ((!a) || (!b) || (!c) || (!d)) {
      fprintf(stderr,"out of memory\n");
      fflush(stderr);
      exit(1);
    }

    a[0] = (sa) ? getmem(abs(sa)) : NULL; if (sa > 0) random_init(a[0], sa);
    b[0] = (sb) ? getmem(abs(sb)) : NULL; if (sb > 0) random_init(b[0], sb);
    c[0] = (sc) ? getmem(abs(sc)) : NULL; if (sc > 0) random_init(c[0], sc);
    d[0] = (sd) ? getmem(abs(sd)) : NULL; if (sd > 0) random_init(d[0], sd);

    for (i=1; i < nrep; i++) {
      if (sa > 0) {a[i] = getmem(sa); random_init(a[i], sa);} else a[i] = a[i-1];
      if (sb > 0) {b[i] = getmem(sb); random_init(b[i], sb);} else b[i] = b[i-1];
      if (sc > 0) {c[i] = getmem(sc); random_init(c[i], sc);} else c[i] = c[i-1];
      if (sd > 0) {d[i] = getmem(sd); random_init(d[i], sd);} else d[i] = d[i-1];
    }

    t = chrono();
    for (i=0; i < nrep; i++) tst->calc(a[i], b[i], c[i], d[i], n, n0);
    t = chrono()-t;

    if (sa) freemem(a[0]);
    if (sb) freemem(b[0]);
    if (sc) freemem(c[0]);
    if (sd) freemem(d[0]);
    for (i = 1; i < nrep; i++) {
      if (sa > 0) freemem(a[i]);
      if (sb > 0) freemem(b[i]);
      if (sc > 0) freemem(c[i]);
      if (sd > 0) freemem(d[i]);
    }
    free(a); free(b); free(c); free(d);

    return(t/nrep);
  }

}

                /* +-------------------------------+
                   |  point d'entre du programme  |
                   +-------------------------------+ */

int main(int argc, char **argv) {

  int temps   = 1000;      /* temps de base en ms       */
  int minsize = 10;        /* plage de test             */
  int maxsize = 100;
  int n0      = 32;        /* seuil pour Karatsuba      */
  int check   = 0;         /* contrle                  */
  int percent = 0;         /* affichage relatif  kara0 */
  int help    = 0;         /* affiche l'aide            */
  double tk = 0, t;        /* temps mesurs             */
  int i,n, rc=0;

  srand(time(NULL));

  /* analyse la ligne de commande */
  argc--; argv++;
  for (; argc; argc--, argv++) {

    if ((*argv)[0] != '-') {fprintf(stderr,"paramtre non reconnu : %s\n",*argv); rc = 1;}
    else {

      for (i=0; (i < nbtests) && (strcmp(argv[0]+1, test[i].name)); i++);
      if (i < nbtests) {test[i].sel ^= 1;}
      else {

        if      (!strcmp(*argv,"-time"))   {argv++; argc--; temps   = atoi(*argv);}
        else if (!strcmp(*argv,"-min"))    {argv++; argc--; minsize = atoi(*argv);}
        else if (!strcmp(*argv,"-max"))    {argv++; argc--; maxsize = atoi(*argv);}
        else if (!strcmp(*argv,"-n"))      {argv++; argc--; maxsize = atoi(*argv); minsize=maxsize;}
        else if (!strcmp(*argv,"-n0"))     {argv++; argc--; n0      = atoi(*argv);}
        else if (!strcmp(*argv,"-seed"))   {argv++; argc--; srand(atoi(*argv));}

        else if (!strcmp(*argv,"-check"))  {check   = 1;}
        else if (!strcmp(*argv,"-percent")){percent = 1;}
        else if (!strcmp(*argv,"-h"))      {help    = 1;}

        else if (!strcmp(*argv,"-all"))    {for (i=0; i < nbtests; test[i++].sel ^= 1);}

        else {fprintf(stderr,"option non reconnue : %s\n",*argv); rc = 1;}
      }
    }

    if ((help) || (rc)) {
      fprintf(stderr,"syntaxe : tests [-h] [-check] [-percent] [-seed s] [-time t]\n");
      fprintf(stderr,"                [-min a] [-max b] [-n n] [-n0 n0]\n");
      fprintf(stderr,"                [liste de tests] [-all]\n");
      for (i=0; i < nbtests; i++) {
        fprintf(stderr,"  %-6s: %s\n", test[i].name, test[i].desc);
      }
      fflush(stderr);
      exit(rc);
    }
  }

  /* percent => kara0 */
  if (percent) {test[0].sel = 1;}

  /* affiche les paramtres pour contrle */
  printf("n0   = %d\n",n0);
  printf("min  = %d\n",minsize);
  printf("max  = %d\n",maxsize);
  printf("time = %d\n",temps);
  printf("  n  ");
  for (i=0; i < nbtests; i++) if (test[i].sel) printf(" %6s",test[i].name);
  printf("\n");

  /* excute les tests demands */
  for (n=minsize; n <= maxsize; n++) {
    printf("%4d,", n);
    for (i=0; i < nbtests; i++) if (test[i].sel) {
      t = do_test(test+i, n, n0, temps, check);
      if (check) printf(" %6s","ok");
      else {
        if (i == 0) tk = t; else if (percent) t /= tk;
        printf(" %6.3f",t);
      }
    }
    printf("\n");
  }

  fflush(stdout);
  return(0);
}
